// NOTE: don't use stdcall on INT handlers because
// FPC inserts invisible pushes and pops ...



procedure ProcessIRQ( irq_n:DWord; 
                      p_frame:PZeroRingStackFrame ); stdcall; alias : 'ProcessIRQ';
var
    rw:DWord;
    n:Byte;
begin

  KU_Print('IRQ=');
  KU_PrintUInt(irq_n);
  
  KU_Print('; CS:EIP (h)=');
  
  KU_PrintHex(p_frame^.cs);
  KU_Print(':');
  KU_PrintHex(p_frame^.eip);
  KU_Print('; EFLAGS (h)=');
  KU_PrintHex(p_frame^.eflags);
  KU_PrintLn('');
    
  rw:=gUserIRQHandlers[irq_n](irq_n);
  
  if (rw and UIH_RET_HANDLE_VM86)=0 
    then exit; // - no need to pass control to VM86.
    
  // Here we need to pass control to the VM86 Task!

  
  if irq_n<8 then n:=8+irq_n
   else n:=$70+irq_n;
   
   
  (*
      
   - check VM
   - if VM=0 then JMP VM86:0
   - if VM=1 then use p_frame to change CS:IP, SP, EFLAGS, change 16-bit stack...
  
  *)
   
   
  if (p_frame^.eflags and EFLAGS_VM)=0 then
  begin
    emul_int86(n);
    
    KU_PrintLn('Jumping to V86...');
    
    asm

      // FAR JUMP PTR16:32
      db $EA
      dd 0
      dw (VM_TSS_DESC_N*8)
    end;
    
    
  end
  else // already in VM task
  begin
    KU_PrintLn('IRQ occurs in V86.');
  
    Dec(p_frame^.esp, 6);
    // push flags, cs, ip
    PWord(p_frame^.esp+(p_frame^.ss shl 4))^:=Word(p_frame^.eip and $FFFF);
    PWord(p_frame^.esp+2+(p_frame^.ss shl 4))^:=p_frame^.cs;
    PWord(p_frame^.esp+4+(p_frame^.ss shl 4))^:=Word(p_frame^.eflags and $FFFF);
    // set new flags, cs, ip
    p_frame^.eflags:=p_frame^.eflags and 
      ( not (EFLAGS_IF or EFLAGS_TF or EFLAGS_AC));
    p_frame^.eip:=PWord(n*4)^ and $FFFF;
    p_frame^.cs:=PWord(n*4+2)^ and $FFFF;
 end;
  
  KU_PrintLn('V86 IRQ Handled.');
  
end;


procedure LowLev_IRQ0_Handler; external name 'LowLev_IRQ0_Handler';
procedure LowLev_IRQ1_Handler; external name 'LowLev_IRQ1_Handler';
procedure LowLev_IRQ2_Handler; external name 'LowLev_IRQ2_Handler';
procedure LowLev_IRQ3_Handler; external name 'LowLev_IRQ3_Handler';
procedure LowLev_IRQ4_Handler; external name 'LowLev_IRQ4_Handler';
procedure LowLev_IRQ5_Handler; external name 'LowLev_IRQ5_Handler';
procedure LowLev_IRQ6_Handler; external name 'LowLev_IRQ6_Handler';
procedure LowLev_IRQ7_Handler; external name 'LowLev_IRQ7_Handler';
procedure LowLev_IRQ8_Handler; external name 'LowLev_IRQ8_Handler';
procedure LowLev_IRQ9_Handler; external name 'LowLev_IRQ9_Handler';
procedure LowLev_IRQ10_Handler; external name 'LowLev_IRQ10_Handler';
procedure LowLev_IRQ11_Handler; external name 'LowLev_IRQ11_Handler';
procedure LowLev_IRQ12_Handler; external name 'LowLev_IRQ12_Handler';
procedure LowLev_IRQ13_Handler; external name 'LowLev_IRQ13_Handler';
procedure LowLev_IRQ14_Handler; external name 'LowLev_IRQ14_Handler';
procedure LowLev_IRQ15_Handler; external name 'LowLev_IRQ15_Handler';