// V86 Monitor, emulation of priveleged instructions etc.


procedure handle_vm86_GP;forward;
procedure handle_vm86_NM;forward;




procedure handle_vm86(exc:Byte);
begin
  // see section 15.1.4 of Intel Manual vol. 3
  
  case exc of  
    EXCEPTION_GP: handle_vm86_GP;
    EXCEPTION_NM: handle_vm86_NM; // math
  else
    begin
      KU_Print('V86 FATAL: Unhandled Exception $');
      KU_PrintHex(exc);
      KU_PrintLn('!');
      asm
        cli
        hlt
      end;
    end;
  end; // case excnum
    
end; // handle_vm86



// -------------------
// Emulates INT instruction, CS:IP must point beyond the INT instruction.
// Must be used outside VM86 Context, modifies gVM_TSS !
procedure emul_int86(intn:Byte);
var
  p_stk:PWord; // - v86 stack pointer
begin
  // get CS:IP pair, push FLAGS, CS, IP, clear IF, TF, AC
  p_stk:=PWord((gVM_TSS.ss shl 4)+gVM_TSS.esp);
  
  Dec(p_stk);
  p_stk^:=gVM_TSS.eflags and $FFFF;
  
  
  Dec(p_stk);
  p_stk^:=gVM_TSS.cs and $FFFF;
  
  
  Dec(p_stk);
  p_stk^:=gVM_TSS.eip and $FFFF;
  
    
  KU_Print('int86 at CS:EIP (h)=');
  KU_PrintHex(gVM_TSS.cs);
  KU_Print(':');
  KU_PrintHex(gVM_TSS.eip);
    
  Dec(gVM_TSS.esp, 6);
  
  gVM_TSS.eflags:=gVM_TSS.eflags and 
    ( not (EFLAGS_IF or EFLAGS_TF or EFLAGS_AC));
    
  gVM_TSS.cs:=PWord(intn*4+2)^;
  gVM_TSS.eip:=PWord(intn*4)^;
  
  
  KU_Print(' int86 to CS:EIP (h)=');
  KU_PrintHex(gVM_TSS.cs);
  KU_Print(':');
  KU_PrintHex(gVM_TSS.eip);
  KU_PrintLn('');  
      
end;

// --------------------------------------------------
// GENERAL PROTECTION Fault Handler for VM86
// --------------------------------------------------


procedure handle_vm86_GP;
var
    p_op:PByte; // instruction caused exception
    i,rem:TUInt;
    
label skp_transfer;    
begin
  p_op:=Pointer(gVM_TSS.eip+(gVM_TSS.cs shl 4));
  //  GPF handling for VM86
  
  // klogmsg('switch');  
  
  // if it's an INT instruction then check its argument
  // VM86_INTARG.
  
  // 1) Check whether it is an INT xx instruction
  // 2) Check for INT*
  // 3) Check for CLI, STI, POPF, PUSHF, IRET.
  
  // TODO: see Intel Manual Vol. 2 for details - several checks must be implemented!!!
  //        for full compatibility!!!
  
  
  (*
    Opcode
    
    CC      INT 3       Interrupt 3 - trap to debugger
    CD ib   INT imm8    
    CE      INTO        Interrupt 4 - if overflow flag is 1 (EFLAGS_OF)
    
    ------    
    
    FA      CLI         Clear IF
    FB      STI         Set IF
    
    9D      POPF        POP from stack 16-bits of EFLAGS
    9C      PUSHF       PUSH 16-bits of EFLAGS into stack
    
    CF      IRET        pop IP, pop CS, pop FLAGS
  
  *)
  
  KU_Print('V86: Emulating instruction...');
  
  case p_op^ of
    $FA: begin  // CLI
           gVM_TSS.eflags:=gVM_TSS.eflags and (not EFLAGS_IF);
           Inc(gVM_TSS.eip);
           KU_PrintLn('CLI');
         end;
         
    $FB: begin // STI
           gVM_TSS.eflags:=gVM_TSS.eflags or EFLAGS_IF;
           Inc(gVM_TSS.eip);
           KU_PrintLn('STI');
         end;

    $9C: begin // PUSHF
           Dec(gVM_TSS.esp, 2);
           PWord(gVM_TSS.esp+(gVM_TSS.ss shl 4))^:=DWord(gVM_TSS.eflags 
             and $FFFF);
             
           Inc(gVM_TSS.eip);
           
           KU_PrintLn('PUSHF');
         end;
         
    $9D: begin // POPF           
           gVM_TSS.eflags:=(gVM_TSS.eflags and $FFFF0000) 
             or PWord(gVM_TSS.esp+(gVM_TSS.ss shl 4))^;
             
           Inc(gVM_TSS.esp, 2);
           Inc(gVM_TSS.eip);
           
           KU_PrintLn('POPF');
         end;
         
    $CF: begin // IRET
    
           gVM_TSS.eflags:=(gVM_TSS.eflags and $FFFF0000) 
             or PWord(gVM_TSS.esp+4+(gVM_TSS.ss shl 4))^;
             
           gVM_TSS.cs:=PWord(gVM_TSS.esp+2+(gVM_TSS.ss shl 4))^;
           
           gVM_TSS.eip:=PWord(gVM_TSS.esp+0+(gVM_TSS.ss shl 4))^;
             
           Inc(gVM_TSS.esp, 6);
           
           KU_Print('IRET to CS:EIP (h)=');
           KU_PrintHex(gVM_TSS.cs);
           KU_Print(':');
           KU_PrintHex(gVM_TSS.eip);
           KU_PrintLn('');
         end;
         
    $CC: begin // INT 3
           KU_PrintLn('INT3');
           Inc(gVM_TSS.eip);
           emul_int86(3);
         end;

    $CE: begin // INTO
           KU_PrintLn('INTO');
           Inc(gVM_TSS.eip);
           
           if (gVM_TSS.eflags and EFLAGS_OF) <>0 then
           begin
             emul_int86(4);
           end;
         end;
         
    $CD: begin // INT imm8
           KU_Print('INT $');
           KU_PrintHex(p_op[1]);
           KU_PrintLn('');
    
           Inc(gVM_TSS.eip, 2);
           
           if p_op[1]=VM86_INTARG then // check for VM86 Exit
           begin
             PM_ChangeIRET_Task( @gGDT[VM_TSS_DESC_N], 
                                 @gExcTSS[13],
                                 SYS_TSS_DESC_N*8,
                                 @gGDT[SYS_TSS_DESC_N] );
                                 
             gIsOnVM86Int:=false; // signal VM86 BIOS call completion.
             goto skp_transfer;
           end;
           
           emul_int86( p_op[1] );
           
     skp_transfer:; // no transfer for virtual vm86->pm gate interrupt
           
           
         end;
           
  else // handle unknown opcode sequence
    begin
           
      KU_PrintLn('FAILED!');
      KU_Print('VM86_FATAL: Unk. opcode seq.= ');
      
      for i:=1 to 8 do
      begin
        KU_PrintHex(p_op^);
        Inc(p_op);
      end;
      
      KU_PrintLn('');
      KU_Print('EDX=');
      KU_PrintHex(gVM_TSS.edx);
      KU_PrintLn('');
        
      KU_PrintLn('ESP=');
      KU_PrintHex(gVM_TSS.esp);
      
      {
      Str(gVM_TSS.ss,si);
      KU_PrintLn('ss='+si);
      
      Str(gVM_TSS.eip,si);
      KU_PrintLn('eip='+si);
      Str(gVM_TSS.cs,si);
      KU_PrintLn('cs='+si);
      
      Str(gVM_TSS.eflags,si);
      KU_PrintLn('eflags='+si);
      }
      
      

      
      i:=GU_Div( gVM_TSS.edx, 8, @rem);
      i:=gVM_TSS.iomap[i];
      
      KU_Print('IOMAP[]=');
      KU_PrintHex(i);
      KU_PrintLn('');
      KU_Print('SYSTEM HALTED.');
      
      asm
        cli
        hlt
      end;
    end;
  end; // case



end;


//////// MATH

procedure handle_vm86_NM;
begin
  KU_PrintLn('math');
  asm
    clts    
    hlt
  end;
end;