unit pmode;
interface
uses gentypes;
// descriptors etc.
type

  PZeroRingStackFrame=^TZeroRingStackFrame;
  TZeroRingStackFrame=packed
  record
    eip,
    cs,
    eflags,
    // if stack switch
    esp,
    ss,
    // if VM86
    es,
    ds,
    fs,
    gs: DWord;
  end;



  PSegDescr=^TSegDescr;
  TSegDescr=packed
    record
     case integer of
     0:
      (limit:Word; //   dw ?    ; 15:00
      base_l:Word; //   dw ?    ; 15:00
      base_m:Byte; //   db ?    ; 23:16
      type_b:Byte; //   db ?    ; 11:08 type, 12-S, 14:13 DPL, 15-P, 
      attr_b:Byte; //   db ?    ; 16:19 limit(19:16), 20-AVL, 21-0, 22-D/B, 23-G
      base_h:Byte); //  db ?    ; 31:24
     1:(firstw:DWord;
       secondw:DWord);
     3:(lo0, lo1,
        lo2, lo3:Word);
    end;

    
  TGDTR=packed
    record
      limit:Word;
      base:DWord;
    end;
    

  PTSS32=^TTSS32;
  TTSS32=packed
    record
      t_link,res1:Word;
      esp0:DWord;
      ss0,res2:Word;
      esp1:DWord;
      ss1,res3:Word;
      esp2:DWord;
      ss2,res4:Word;
      pdbr,eip,eflags,
      eax,ecx,edx,ebx,
      esp,ebp,esi,edi:DWord;
      es,res5,cs,res6,ss,res7,
      ds,res8,fs,res9,gs,res10,
      ldt,res11,TDBGF,iomapoffs:Word;
      
      irm:array[0..31] of Byte; // interrupt redirection map for VME bit 0 in CR4
      iomap:array[0..8191] of Byte; // reserve 65536 ports, 0=accessible, 1=locked
      // not needed if CPL=IOPL, but a compulsury one if VM86     
    end;    

    
  PTSS32_0=^TTSS32_0;// shrinked structure for zero privileged TSSes only!
  TTSS32_0=packed
    record
      t_link,res1:Word;
      esp0:DWord;
      ss0,res2:Word;
      esp1:DWord;
      ss1,res3:Word;
      esp2:DWord;
      ss2,res4:Word;
      pdbr,eip,eflags,
      eax,ecx,edx,ebx,
      esp,ebp,esi,edi:DWord;
      es,res5,cs,res6,ss,res7,
      ds,res8,fs,res9,gs,res10,
      ldt,res11,TDBGF,iomapoffs:Word;
    end;
    

// -- constants --
    
const

    DEF_EFLAGS=$202;// IF (9), res(1)
    DEF_PGATTR=$7; // present, writeable, user
    DEF_PGDIRATTR=$7; // present, writeable, user
    
    EFLAGS_NT=1 shl 14; // Nested Task Flag
    EFLAGS_VM=1 shl 17; // Virtual-8086 Mode
    EFLAGS_IF=1 shl 9; // Interrupt Enable Flag
    EFLAGS_OF=1 shl 11; // Overflow flag
    EFLAGS_CF=1 shl 0;
    EFLAGS_ZF=1 shl 6;
    EFLAGS_TF=1 shl 8; // trap flag
    EFLAGS_AC=1 shl 18; // alignment check
    
    
    CLEAR_EFLAGS=1 shl 1; // when all flags are clear

    SEGTYPE_DPL1=1 shl 5;
    SEGTYPE_DPL2=2 shl 5;
    SEGTYPE_DPL3=3 shl 5;
    SEGTYPE_PRESENT=1 shl 7;
    SEGTYPE_CODEDATA=1 shl 4;   
    SEGTYPE_BUSY=2;
    SEGATTR_AVAIL=1 shl 4; //   ; not used by Intel
    SEGATTR_32BIT=1 shl 6;
    SEGATTR_GRAN=1 shl 7;
// -- when SEGTYPE_CODEDATA is set:

    SEGTYPE_ACCESSED=1;
    SEGTYPE_WRITE=2; // writeable data
    SEGTYPE_READ=2; // readable code
    SEGTYPE_CODE=1 shl 3;
    SEGTYPE_CONFORM=1 shl 2; // conforming code
    SEGTYPE_EXPANDD=1 shl 2; // expand-down data

// --- ready-made values

    DEFSEG_TDATA=SEGTYPE_PRESENT or SEGTYPE_CODEDATA or SEGTYPE_WRITE;
    DEFSEG_TCODE=SEGTYPE_PRESENT or SEGTYPE_CODEDATA or SEGTYPE_READ or SEGTYPE_CODE;
    DEFSEG_ATTR=SEGATTR_32BIT or SEGATTR_GRAN;
    DEFSEG_LIMIT=$FFFFF; // *4KB=4GB limit

// -- system segments (CODEDATA=0):

    SEGTYPE_LDT=$2;//0010;
    SEGTYPE_TASKGATE=$5;//0101;
    SEGTYPE_TSS32=$9;//0b1001;
    // bit 9 in the second DWord means BUSY flag for TSS32
    SEGTYPE_CALLGATE32=$C;//0b1100;
    SEGTYPE_INTGATE32=$E;//0b1110;
    SEGTYPE_TRAPGATE32=$F;//0b1111;

// -- functions --

  procedure PM_SetDescr( p_d:PSegDescr;
                         base,
                         limit,
                         type_b,
                         attr_b:DWord );
                         
  procedure PM_SetIntDescr( p_d:PSegDescr;
                            sel:Word;
                            proc_offset,
                            type_b:DWord );
                            
 (* Interrupt Gate
  4  31..16    15 13..13 12..8  7..5 4..0
     offset    P   DPL   0D110  000   
    
  0  31..16  15..0
     seg_sel  offset
 
 *)
                         

                         
  procedure PM_SetSelPresent( p_d:PSegDescr;
                              present_flag:Boolean );

                              
  function PM_GetSelPresent( p_d:PSegDescr ):Boolean;


  procedure PM_SetBusy( p_d:PSegDescr );
  
  procedure PM_ClearBusy( p_d:PSegDescr );
  
  
  procedure PM_ChangeIRET_Task( p_prev_desc:PSegDescr;
                                p_cur_tss:PTSS32;
                                next_tsel:Word; // next task selector
                                p_next_desc:PSegDescr ); // next task descriptor
  (*
    - changes t_link field of the current task (p_cur_tss)
    and Busy flags of the previous (p_prev_desc) and the new descriptors
    (p_next_desc).
    
    NT flag of the current task must be set and a task switch to 'next_tsel'
    can be done via IRET instruction.

   e.g. This function may be used in #GPF handlers to pass control back to
   PM 32-bit tasks when encountering invalid VM86 code etc.
  *)
                               
                            
  function PM_GetDescrBase(p_desc:PSegDescr):DWord;

  
  
implementation

  procedure PM_SetBusy( p_d:PSegDescr );
  begin
    p_d^.secondw:= p_d^.secondw or $200; // bit 9
  end;

  
  procedure PM_ClearBusy( p_d:PSegDescr );
  begin
    p_d^.secondw:= p_d^.secondw and $FFFFFDFF; // clear bit 9
  end;



  procedure PM_SetSelPresent( p_d:PSegDescr;
                              present_flag:Boolean );
  begin
    p_d^.type_b:=(p_d^.type_b and $7F) or (Ord(present_flag) shl 7);
  end;

  
  function PM_GetSelPresent( p_d:PSegDescr ):Boolean;
  begin
    PM_GetSelPresent:=false;
    
    if (p_d^.type_b shr 7) > 0
      then PM_GetSelPresent:=true;
  end;

  
  
  procedure PM_SetDescr( p_d:PSegDescr;
                         base,
                         limit,
                         type_b,
                         attr_b:DWord );

  begin
    p_d^.base_l:= base and $FFFF;
    p_d^.base_m:= (base shr 16) and $FF;
    p_d^.base_h:= (base shr 24) and $FF;
  
    p_d^.limit:= limit and $FFFF;
    p_d^.attr_b:= ((limit shr 16) and $F) or attr_b;
    p_d^.type_b:= type_b;
  end;
  
  
  procedure PM_SetIntDescr( p_d:PSegDescr;
                            sel:Word;
                            proc_offset,
                            type_b:DWord );

 (* Interrupt Gate
  4  31..16    15 13..13 12..8  7..5 4..0
     offset    P   DPL   0D110  000

  0  31..16  15..0
     seg_sel  offset

 *)
 begin
   PM_SetDescr( p_d, sel, 0, type_b, 0 );
   
   p_d^.lo0:=proc_offset and $FFFF;
   p_d^.lo3:=proc_offset shr 16;   
 end;
 
 
 
 
  procedure PM_ChangeIRET_Task( p_prev_desc:PSegDescr;
                                p_cur_tss:PTSS32;
                                next_tsel:Word; // next task selector
                                p_next_desc:PSegDescr ); // next task descriptor
  begin
    PM_ClearBusy( p_prev_desc );
    PM_SetBusy( p_next_desc );
    
    p_cur_tss^.t_link:= next_tsel;
    
    // NT flag must be set!
    // and now executing IRETD will result in a switch to the 'next_tsel' task.
  end;
  
  
  function PM_GetDescrBase(p_desc:PSegDescr):DWord;
  begin
    PM_GetDescrBase:=p_desc^.base_l or (p_desc^.base_m shl 16) 
            or (p_desc^.base_h shl 24);
  end;

  


end.

(*

macro def_ercode_table  ; 0=no eror code, 1=error code in stack
{
 ercode_table   db 0,0,0,0,0,0,0,0
        db 1,0,1,1,1,1,1,0
        db 0,1,0,0,0,0,0,0
        db 0,0,0,0,0,0,0,0
 ; [esp]=errcode (32-bit) ... see 5.11 ERROR CODE section in Intel's vol.3
}

*)
