{$M 2048,2048} // stack size, heap size in ELF
// in FPC 1.0.10 minimum heap size must be > 1024
// now kernel doesn't use its heap but if needed this value
// must be increased. the stack specified by the RTL is also not
// used!


{$ASMMODE intel}
uses gentypes, genutils, sdconsts, sdcommon, pmode, kutils, sd32_sdk,
     sdvm86task, sdexcep, sdintr, sdmath, sdcontexts;

const
  k_stack_size=65536;

var 
  ktdtr:TGDTR; // GDTR or IDTR
  i:TUInt;
  p:Pointer;
  
  k_stack:array[0..k_stack_size-1] of Byte;


// ---------

  procedure DisplayLogo;
  begin
    KU_PrintLn(SDVERSTR);
    KU_PrintLn(SDCONTACTSTR);
    KU_PrintLn('-=-=-=-=-=-=-=-=-=-=');
    KU_PrintLn(SDGPLSTR);
  end;

//-------------------------------------
 
// KERNEL START: 
 
begin
  // setup GDT
  // GDT, TSS
  PM_SetDescr(@gGDT[0],0,0,0,0);
  PM_SetDescr(@gGDT[1],0,DEFSEG_LIMIT,DEFSEG_TCODE,DEFSEG_ATTR);
  PM_SetDescr(@gGDT[2],0,DEFSEG_LIMIT,DEFSEG_TDATA,DEFSEG_ATTR);
  
  PM_SetDescr(@gGDT[3],0,DEFSEG_LIMIT,DEFSEG_TCODE or SEGTYPE_DPL3,DEFSEG_ATTR);
  PM_SetDescr(@gGDT[4],0,DEFSEG_LIMIT,DEFSEG_TDATA or SEGTYPE_DPL3,DEFSEG_ATTR);
 
  ktdtr.limit:=sizeof(gGDT)-1;
  ktdtr.base:=DWord(@gGDT[0]);

  // now load GDTR and segments
 
  asm
    //mov ecx,esp
    lea ecx,k_stack
    add ecx,k_stack_size-4
    
    
    
    push word 8 // push FAR return address into stack, first CS
    lea eax,@cont
    push eax // now EIP
  
    lea eax,ktdtr
    lgdt [eax]
    retf  
  @cont:
    // here CS=new CS, now fill in the rest of seg. registers
    mov ax,16
    mov ss,ax
    mov esp,ecx 
    mov ds,ax
    mov es,ax
    mov ax,0
    mov fs,ax
    mov gs,ax
            
  end ['EAX','ECX'];

  // now GDT and segments ok 
  // let's setup TSSes, interrupts, exceptions...
  
  
  DisplayLogo;
  
   
  FillByte(gIDT,sizeof(gIDT),0);
  
  // exceptions

  InitExcHandlers;

  // system TSS

  PM_SetDescr(@gGDT[SYS_TSS_DESC_N], DWord(@gSysTSS), sizeof(TTSS32_0)-1,
    SEGTYPE_TSS32 or SEGTYPE_PRESENT, 0);
              
  // IDTR

  ktdtr.limit:=sizeof(gIDT)-1;
  ktdtr.base:=DWord(@gIDT[0]);

 // load TR
  asm
    mov eax, SYS_TSS_DESC_N*8
    ltr ax
  // load IDTR
    lea eax,ktdtr
    lidt [eax]
  end ['EAX'];
  
  
  // initialize contexts

  InitSDContexts;
    
    
  // initialize IRQs

  InitIntHandlers;
  // - enables NMIs, makes handlers be ready for STI
  
  
        
  // old LIDT was here...  
     
  // -------------------------------------------------
  // INITIALIZE VM86 TASK and its code:
  // VM_86

  InitVM86;


  // --- END PREPARING VM86 ----
  
  // INITIALIZE math exceptions handling...
  
  InitSDMath;

  
    
  // Now enable interrupts
  
  // ------------------------------------------------------
  
  // enable scheduler
  
  // TODO: initialize drivers, run command interpreter
  
  // TODO: initialize VM86 here
  

 
  
  
  // ...
{  
  gVM_TSS.eax:=$13;
  
  emul_int86($10);
    asm

      // FAR JUMP PTR16:32
      db $EA
      dd 0
      dw (VM_TSS_DESC_N*8)
    end;
    
 } 
 
  DumpTSS( VM_TSS_DESC_N*8, @gVM_TSS );
  
  asm
   sti
   
  @a:
   mov eax,$B8000;
   mov byte [eax],'L'
   not byte [eax+1]  
   hlt
   wait // provoke math exc.
   
   jmp @a
  end;
 

end.