unit sdcommon;
interface
{$UNITPATH sdk}
uses gentypes, sdconsts, pmode, sd32_sdk;
type

  // Stack frame for ProcessSysInt:

  PSysIntStackFrame=^TSysIntStackFrame;
  TSysIntStackFrame=packed
  record
    // We may modify this fields
    // to make the corresponding registers
    // to be changed because the system
    // interrupt handler loads them via POPAD.
    
    // After PUSHAD instruction:
    edi,
    esi,
    ebp,
    esp,
    ebx,
    edx,
    ecx,
    eax,
    // After INT instruction:
    eip,
    cs,
    eflags,
    // On stack switch:
    old_esp,
    ss:DWord;
  end;

var
  gMB_EAX:DWord;external name '_EAX_MB';
  gMB_EBX:DWord;external name '_EBX_MB';
  
  gGDT:array[0..GDTDESC_COUNT-1] of TSegDescr;
  gIDT:array[0..IDTDESC_COUNT-1] of TSegDescr;
    
  gSysTSS:TTSS32_0; // main TSS
                        

  procedure StopKernel; // stops execution - used in critical situations.
  
  procedure DumpTSS( trval:Word; p_tss:PTSS32 );
  

implementation
uses kutils;


  procedure DumpTSS( trval:Word; p_tss:PTSS32 );
  begin
    KU_PrintLn('');
    KU_Print('TSS Dump in Hex: TSS Sel=');
    KU_PrintHex( trval );
    KU_PrintLn('');
    KU_Print('CS:EIP=');
    
    KU_PrintHex( p_tss^.cs );
    KU_Print(':');
    KU_PrintHex( p_tss^.eip );
    
    KU_Print('  SS:ESP=');
    KU_PrintHex( p_tss^.ss );
    KU_Print(':');
    KU_PrintHex( p_tss^.esp );
    
    KU_Print('  SS0:ESP0=');
    KU_PrintHex( p_tss^.ss0 );
    KU_Print(':');
    KU_PrintHex( p_tss^.esp0 );
    
    
    KU_Print('  EFLAGS=');
    KU_PrintHex( p_tss^.eflags );
    
    
    KU_PrintLn('');
    
    KU_Print('DS=');
    KU_PrintHex( p_tss^.ds );
    
    KU_Print('  ES=');
    KU_PrintHex( p_tss^.es );
    
    KU_Print('  FS=');
    KU_PrintHex( p_tss^.fs );
    
    KU_Print('  GS=');
    KU_PrintHex( p_tss^.gs );
    
    KU_Print('  LDTR=');
    KU_PrintHex( p_tss^.ldt );
        
    KU_PrintLn('');
    
    KU_Print('PDBR=');
    KU_PrintHex( p_tss^.pdbr );

    KU_Print('  TDBGF=');
    KU_PrintHex( p_tss^.TDBGF );
    
    KU_Print('  t_link=');
    KU_PrintHex( p_tss^.t_link );
    
    KU_PrintLn('');    
    
    KU_Print('EBP=');
    KU_PrintHex( p_tss^.ebp );
    
    KU_Print('  ESI=');
    KU_PrintHex( p_tss^.esi );
    
    KU_Print('  EDI=');
    KU_PrintHex( p_tss^.edi );
    
    KU_PrintLn('');    
    
    KU_Print('EAX=');
    KU_PrintHex( p_tss^.eax );

    KU_Print('  EBX=');
    KU_PrintHex( p_tss^.ebx );
    
    KU_Print('  ECX=');
    KU_PrintHex( p_tss^.ecx );
    
    KU_Print('  EDX=');
    KU_PrintHex( p_tss^.edx );
    
    KU_PrintLn('');
    
  end;

  procedure StopKernel;
  begin
    KU_Print('Kernel Halt.');
    while true do
    begin
      asm
        cli
        hlt
      end;
    end; // while
  end; // StopKernel



end.