// TODO: place scheduler code here!


unit sdcontexts;
interface
uses gentypes,pmode;

type

  TSDWakeMethod=(sdwmNone, sdwmOnSkipped, sdwmOnTime);
   // - sdwmNone - the context will be awaken by another context via API.
   // - sdwmOnSkipped - the context sets itself into a sleeping state
   // informing that it is idle and the scheduler puts him immediately
   // aside and runs the next awaiting context and mark the sleeping one
   // as "ALREADY SKIPPED" so the next time the scheduler encounters
   // that context, it will be woken up if sdwmOnSkipped is set.
   // - sdwmOnTime - a period is set when the context must be awaken.   

  PSD_TSS_TR=^TSD_TSS_TR;
  TSD_TSS_TR=
  record
    p_tss:PTSS32;
    trval:Word;  
  end;

  PSDContext=^TSDContext;
  TSDContext=
  record
    tss_tr:TSD_TSS_TR;
    is_ring0:Boolean; // if true then p_tss=PTSS32_0
    is_vm86:Boolean; // if used to emulate 8086
    is_read_only:Boolean; // kernel main TSS and main VM86 are read-only
    fpu_ctx:array[0..107] of Byte; // 108 bytes for FPU context
    /// execution info
    is_sleeping:Boolean;
     wake_method:TSDWakeMethod;
     is_skipped:Boolean; // sdwmOnSkipped
     wake_time:TUInt;
    process_id:TUInt; 
      // - a process is characterized by having a separated addr. 
      // space and being loaded from an EXE.
  end;
  
  
  
var
  gSysCtxID:TUInt;
  gVM86CtxID:TUInt;
  
  
 
//////////////////////////////////////////////////
/////////
/////////////
 
  procedure InitSDContexts;
   // - initializes gSysCtxID, gVM86CtxID
  
  function GetContextFromID( ctx_id:TUInt ):PSDContext;
  
  function GetCtxIDfromTR( trval:Word; p_out_ctx_id:PUInt ):Boolean;


  
/////////////////////////////////////////////////////////////
// IMPLEMENTATION /////////////////////////////
/////////////////////////////////////////////////////////////
  
implementation
uses gulnklst, sdconsts, sdcommon, sdvm86task;

const
  CONTEXTS_AMOUNT=(USER_TSS_COUNT+2);

type
  PContextsEntry=^TContextsEntry;
  TContextsEntry=
  record
    hdr:TGLinkHdr;
    ctx:TSDContext;
  end;

var

 // use linked lists
  g_contexts_ldata:array[0..CONTEXTS_AMOUNT-1] of TContextsEntry;
  g_ctxll:TLLDesc;
   // - 0 is for the main kernel, 1 is for the main VM86,
   // - the rest is free.
  g_user_tss:array[0..USER_TSS_COUNT-1] of TTSS32; // system/user contexts
  
  
  
 //////////////////////
 /////// FUNCTIONS
 //////////////////////
 
 function GetContextFromID( ctx_id:TUInt ):PSDContext;
 begin
   exit( PSDContext( LnkLst_OfsToPtr( @g_ctxll, ctx_id ) ) );
 end;
 
 
 function GetCtxIDfromTR( trval:Word; p_out_ctx_id:PUInt ):Boolean;
 var
   e:PContextsEntry;
   nofs:TUInt;
 begin
   nofs:= g_ctxll.alloc_items.ofs_first;
   repeat     
     e:= LnkLst_OfsToPtr( @g_ctxll, nofs );
     
     
     if e^.ctx.tss_tr.trval = trval then
       begin
         p_out_ctx_id^:=nofs;
         exit( true );
       end;
     
     nofs:= e^.hdr.ofs_next;
     
   until nofs = g_ctxll.alloc_items.ofs_first;
   
   exit( false );   
   
 end; // end of GetCtxIDfromTR
   
   
  // INITIALIZATION ROUTINE. MUST NOT INVOLVE EXCEPTIONS OR INTERRUPTS!
  
  procedure InitSDContexts;
  var
    i:TUInt;
    e:TContextsEntry;
  begin

    // initialize linked list
    
    FillByte( g_contexts_ldata, sizeof( g_contexts_ldata ), 0 );
    
    LnkLst_Init( @g_ctxll, sizeof(TContextsEntry), 
        CONTEXTS_AMOUNT, @g_contexts_ldata );
        
    // reserve first two entries for main kernel and VM86
    
    // add kernel tss:
    
    FillByte( e, sizeof( e ), 0 );
    
    e.ctx.tss_tr.p_tss:=@gSysTSS;
    e.ctx.tss_tr.trval:=SYS_TSS_DESC_N*8;
    e.ctx.is_ring0:=true;
    e.ctx.is_vm86:=false;
    e.ctx.is_read_only:=true;
        
    LnkLst_Add( @g_ctxll, @e, @gSysCtxID );
    
    // add vm86
    
    e.ctx.tss_tr.p_tss:=@gVM_TSS;
    e.ctx.tss_tr.trval:=VM_TSS_DESC_N*8;
    e.ctx.is_ring0:=false;
    e.ctx.is_vm86:=true;
    e.ctx.is_read_only:=true;
    
    LnkLst_Add( @g_ctxll, @e, @gVM86CtxID );
    
    // setup descriptors and mark as NON-PRESENT
    
    FillByte( g_user_tss, sizeof( g_user_tss ), 0 );
    
    
    for i:=0 to USER_TSS_COUNT-1 do
    begin
      PM_SetDescr( @gGDT[i+USER_TSS_DESC_N], DWord(@g_user_tss[i]), sizeof(TTSS32)-1,
         SEGTYPE_TSS32, 0); //  without SEGTYPE_PRESENT
    end;
    
       
  end; // end of InitSDContexts

end.