unit sdexcep;
interface
uses gentypes,sdconsts,sdcontexts, pmode;

type
  TUserExcHandler=function( exc_num:TUint; ctx_id:TUInt ):Boolean;
  // context IDs are used because when a fatal exception happens
  // the invalid contexts may be killed and the array and pointers
  // may change.


var
  gExcStacks:array[0..(EXCEPTIONS_COUNT)*EXC_STACK_SZ-1] of Byte;
  gExcTSS:array[0..EXCEPTIONS_COUNT-1] of TTSS32_0;


// --------------- PROCEDURES ------------


  procedure InitExcHandlers;

  function AddUserExcHandler( hproc:TUserExcHandler;
                              var handler_id:TUInt ):Boolean;

  procedure RemoveUserExcHandler( handler_id:TUInt );
  
  

  
  procedure ExcDispatch(trval:DWord); stdcall;   
  // - used by asm code.


implementation
uses gulnklst, sdcommon, kutils;
type
  PExcHandlersEntry=^TExcHandlersEntry;
  TExcHandlersEntry=
  record
    hdr:TGLinkHdr;
    handler:TUserExcHandler;
  end;
  
var
  exc_handlers:array[0..EXCHANDLERS_MAXCOUNT-1] of TExcHandlersEntry;
  exc_handlers_list:TLLDesc;



  // -----------------------------------------------
  // Generic exception handler - TSS main code
  // -----------------------------------------------
  procedure ExcHandler; external name 'ExcHandler';
  {$L exc_a.o}

  procedure HandleException( excnum:Byte; trval:Word ); // exception or INTerrupt (IRQ) number
  var
      p_op:PByte;
      i, ofs_cur:TUInt;
      handled:Boolean;
      p_hnde:PExcHandlersEntry;
      tss_tr:TSD_TSS_TR;
      ctx_id:TUInt;

  begin
  
    (*
       All exceptions caused by running V86 task are sent to
       specialized v86 exception handler and are not processed here.
    *)

    // TODO: - math exception handler, save/restore FPU context...
    
    
    
    tss_tr.trval:=gExcTSS[excnum].t_link;
    
    if not GetCtxIDfromTR( tss_tr.trval, @ctx_id ) then
    begin
      // TR is not associated with any context, BUG!
      
      KU_PrintLn('');
      KU_PrintLn('FATAL: TR is not associated with any context!!!');
      KU_Print('TR= $');
      KU_PrintHex( trval );
      
      StopKernel;      
    end;
    
    
    tss_tr.p_tss:=PTSS32( PM_GetDescrBase( @gGDT[tss_tr.trval shr 3] ) );

    // cycle:
    
    i:=exc_handlers_list.alloc_items.links_count;
    
    ofs_cur:=exc_handlers_list.alloc_items.ofs_first;
    
    handled:=false;

    
    while i>0 do
    begin
      p_hnde:=LnkLst_OfsToPtr( @exc_handlers_list, ofs_cur );
      
      if p_hnde^.handler( excnum, ctx_id ) then
      begin
        handled:=true;
      end;
      
      ofs_cur:=p_hnde^.hdr.ofs_next;
      Dec(i);
    end;

    if handled then exit;


    KU_Print('Unhandled exception $');
    KU_PrintHex(excnum);
    
    DumpTSS( tss_tr.trval, tss_tr.p_tss );

    p_op:=Pointer(tss_tr.p_tss^.eip);

    KU_Print('Opcodes (hex)= ');

    for i:=1 to 8 do
    begin
      KU_PrintHex(p_op^);
      KU_Print(' ');
      Inc(p_op);
    end;

    KU_PrintLn('');
    
    StopKernel;
    // HALT.


  end; // HandleException



  procedure SetupExceptions;
  var i:TUInt;
  begin
    for i:=0 to EXCEPTIONS_COUNT-1 do
    begin
      gExcTSS[i].esp:=DWord(@gExcStacks[i*EXC_STACK_SZ+EXC_STACK_SZ]);
      gExcTSS[i].eip:=DWord(@ExcHandler);
      gExcTSS[i].cs:=8;
      gExcTSS[i].ds:=16;
      gExcTSS[i].es:=16;
      gExcTSS[i].fs:=16;
      gExcTSS[i].gs:=16;
      gExcTSS[i].ss:=16;
      gExcTSS[i].eflags:=0;
      gExcTSS[i].pdbr:=0;

      PM_SetDescr(@gIDT[i], (EXC_TSS_DESC_N+i)*8, 0,
        SEGTYPE_TASKGATE or SEGTYPE_PRESENT,0);

      PM_SetDescr(@gGDT[i+EXC_TSS_DESC_N], DWord(@gExcTSS[i]),
        sizeof(TTSS32_0)-1,SEGTYPE_TSS32 or SEGTYPE_PRESENT,0);
    end; // for i
  end; // SetupExceptions


// ------- second low-level -------

  procedure ExcDispatch(trval:DWord); stdcall; alias : 'ExcDispatch';
  begin
    trval:=trval div 8;
    if (trval >= EXC_TSS_DESC_N)
      and (trval < EXC_TSS_DESC_N+EXCEPTIONS_COUNT) then
    begin
      HandleException(trval-EXC_TSS_DESC_N, trval);
      exit;
    end;
    KU_DispStr(0,'FATAL: IVALID TR VALUE FOR EXCEPTION! KERNEL BUG!');
    
    StopKernel;
  end; // ExcDispatch

// -------- initialization and user handlers ------
  
  
  procedure InitExcHandlers;
  begin
    SetupExceptions;
    
    LnkLst_Init( @exc_handlers_list, sizeof(TExcHandlersEntry), 
        EXCHANDLERS_MAXCOUNT, @exc_handlers );       

  end;
  

  function AddUserExcHandler( hproc:TUserExcHandler;
                              var handler_id:TUInt ):Boolean;
  var
    hnde:TExcHandlersEntry;
  begin
    hnde.handler:=hproc;
  
    AddUserExcHandler:=
      LnkLst_Add( @exc_handlers_list, @hnde, @handler_id);  
  end;

  
  procedure RemoveUserExcHandler( handler_id:TUInt );
  begin
    LnkLst_Remove( @exc_handlers_list, handler_id );      
  end;
  
end.