unit sdintr;
interface
uses gentypes, sdconsts, pmode;

type
  TUserIRQHandler=procedure( irq_n:TUInt; vm86:Boolean; p_frame:PZeroRingStackFrame );
  // - irq_n = 0..15, vm86=true if VM flag is set
var
  gShouldSendEOI:Boolean;
   // - used by vm86 to skip OUT $20,$20 ...
   // always set to TRUE before calling chained user handlers!!!


  procedure InitIntHandlers;
  // - enables NMIs, makes handlers be ready for STI


  function AddUserIRQHandler( hproc:TUserIRQHandler;
                              var handler_id:TUInt ):Boolean;

  procedure RemoveUserIRQHandler( handler_id:TUInt );


// used by asm code: 
  
  procedure ProcessIRQ( irq_n:DWord;
                        p_frame:PZeroRingStackFrame ); stdcall;
  
  

implementation
uses gulnklst, kutils, sdcommon, sdsysint;
type
  PIntHandlersEntry=^TIntHandlersEntry;
  TIntHandlersEntry=
  record
    hdr:TGLinkHdr;
    handler:TUserIRQHandler;
  end;
  
var
  int_handlers:array[0..IRQHANDLERS_MAXCOUNT-1] of TIntHandlersEntry;
  int_handlers_list:TLLDesc;




  procedure LowLev_IRQ0_Handler; external name 'LowLev_IRQ0_Handler';
  procedure LowLev_IRQ1_Handler; external name 'LowLev_IRQ1_Handler';
  procedure LowLev_IRQ2_Handler; external name 'LowLev_IRQ2_Handler';
  procedure LowLev_IRQ3_Handler; external name 'LowLev_IRQ3_Handler';
  procedure LowLev_IRQ4_Handler; external name 'LowLev_IRQ4_Handler';
  procedure LowLev_IRQ5_Handler; external name 'LowLev_IRQ5_Handler';
  procedure LowLev_IRQ6_Handler; external name 'LowLev_IRQ6_Handler';
  procedure LowLev_IRQ7_Handler; external name 'LowLev_IRQ7_Handler';
  procedure LowLev_IRQ8_Handler; external name 'LowLev_IRQ8_Handler';
  procedure LowLev_IRQ9_Handler; external name 'LowLev_IRQ9_Handler';
  procedure LowLev_IRQ10_Handler; external name 'LowLev_IRQ10_Handler';
  procedure LowLev_IRQ11_Handler; external name 'LowLev_IRQ11_Handler';
  procedure LowLev_IRQ12_Handler; external name 'LowLev_IRQ12_Handler';
  procedure LowLev_IRQ13_Handler; external name 'LowLev_IRQ13_Handler';
  procedure LowLev_IRQ14_Handler; external name 'LowLev_IRQ14_Handler';
  procedure LowLev_IRQ15_Handler; external name 'LowLev_IRQ15_Handler';

{$L intr_a.o}


  procedure SetLowLevIntHandlers;
  begin
    PM_SetIntDescr(@gIDT[$20], 8, DWord(@LowLev_IRQ0_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$21], 8, DWord(@LowLev_IRQ1_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$22], 8, DWord(@LowLev_IRQ2_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$23], 8, DWord(@LowLev_IRQ3_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$24], 8, DWord(@LowLev_IRQ4_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$25], 8, DWord(@LowLev_IRQ5_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$26], 8, DWord(@LowLev_IRQ6_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$27], 8, DWord(@LowLev_IRQ7_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$28], 8, DWord(@LowLev_IRQ8_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);


    PM_SetIntDescr(@gIDT[$29], 8, DWord(@LowLev_IRQ9_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$2A], 8, DWord(@LowLev_IRQ10_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$2B], 8, DWord(@LowLev_IRQ11_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$2C], 8, DWord(@LowLev_IRQ12_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);


    PM_SetIntDescr(@gIDT[$2D], 8, DWord(@LowLev_IRQ13_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$2E], 8, DWord(@LowLev_IRQ14_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);

    PM_SetIntDescr(@gIDT[$2F], 8, DWord(@LowLev_IRQ15_Handler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT);


   // ---

    PM_SetIntDescr(@gIDT[SYS_INT_N], 8, DWord(@SysIntHandler),
      SEGTYPE_INTGATE32 or SEGTYPE_PRESENT or SEGTYPE_DPL3);

  end; // SetLowLevIntHandlers
  
  
  procedure SetupPIC;
  begin
    KU_OutP($20,$11); // start 8259 initialization 8259 Master=$20, Slave=$A0
    KU_OutP($A0,$11);
    KU_OutP($21,$20); // master base interrupt vector
    KU_OutP($A1,$28); // slave base interrupt vector
  
    KU_OutP($21, 1 shl 2); // bitmask for cascade on IRQ2
  
    KU_OutP($A1,2); //cascade on IRQ2
    KU_OutP($21,1);// finish 8259 initialization
    KU_OutP($A1,1);
  
    //
  
    KU_OutP($21,0); // enable all IRQs at Master
    KU_OutP($A1,0); // enable all IRQs at Slave
  
    // Enable NMI
    KU_OutP($70, KU_InP($70) and $7F);
  end;
  
  
  
  procedure InitIntHandlers;
  begin
    SetLowLevIntHandlers;
    SetupPIC;
    
    LnkLst_Init( @int_handlers_list, sizeof(TIntHandlersEntry), 
       IRQHANDLERS_MAXCOUNT, @int_handlers );       
    
  end;
  
  
  
  function AddUserIRQHandler( hproc:TUserIRQHandler;
                              var handler_id:TUInt ):Boolean;
  var
    hnde:TIntHandlersEntry;
  begin
    hnde.handler:=hproc;
  
    AddUserIRQHandler:=
      LnkLst_Add( @int_handlers_list, @hnde, @handler_id);  
  end;

  
  
  procedure RemoveUserIRQHandler( handler_id:TUInt );
  begin
    LnkLst_Remove( @int_handlers_list, handler_id );
  end;
  

  procedure ProcessIRQ( irq_n:DWord;
                        p_frame:PZeroRingStackFrame ); stdcall; alias : 'ProcessIRQ';
  var
      n, ofs_cur:TUInt;
      is_vm:Boolean;
      p_hnde:PIntHandlersEntry;
  begin
  
    gShouldSendEOI:=true; // planned to be used by the vm86 emulator
    
    
    is_vm:=(p_frame^.eflags and EFLAGS_VM)<>0;
    
    n:=int_handlers_list.alloc_items.links_count;
    
    ofs_cur:=int_handlers_list.alloc_items.ofs_first;
    
    while n>0 do
    begin
      p_hnde:=LnkLst_OfsToPtr( @int_handlers_list, ofs_cur );
      
      p_hnde^.handler( irq_n, is_vm, p_frame );
      
      ofs_cur:=p_hnde^.hdr.ofs_next;
      Dec(n);
    end;
        
    if gShouldSendEOI then KU_OutP($20,$20);

  end; // - ProcessIRQ



end.